@@ -7,10 +7,10 @@
 ;;         Damien Picard
 ;; Maintainer: Damien Picard
 ;; URL: https://github.com/Kungsgeten/ink-mode
-;; Package-Version: 0.3.2
-;; Package-Revision: 71d215712067
+;; Package-Version: 20260206.923
+;; Package-Revision: 04e141fbba63
 ;; Keywords: languages, wp, hypermedia
-;; Package-Requires: ((emacs "26.1"))
+;; Package-Requires: ((emacs "29.1"))
 
 ;; This file is not part of GNU Emacs.
 
@@ -39,14 +39,15 @@
 
 ;;; Code:
 
-(require 'rx)
 (require 'comint)
 (require 'thingatpt)
+(require 'imenu)
 (require 'outline)
 (require 'subr-x)
 (require 'easymenu)
 (require 'flymake)
-(require 'seq)
+(require 'cl-lib)
+(require 'rx)
 
 (defgroup ink nil
   "Major mode for writing interactive fiction in Ink."
@@ -60,28 +61,18 @@
   :group 'ink
   :group 'faces)
 
-(defvar ink-mode-hook nil)
+(defvar-keymap ink-mode-map
+  :doc "Keymap for `ink-mode'."
+  "C-c C-c" #'ink-play
+  "C-c C-p" #'ink-play-knot
+  "C-c C-o" #'ink-follow-link-at-point
+  "C-c C-h" #'ink-display-manual)
 
-(defvar ink-mode-map
-  (let ((map (make-keymap)))
-    (define-key map (kbd "C-c C-c") 'ink-play)
-    (define-key map (kbd "C-c C-p") 'ink-play-knot)
-    (define-key map (kbd "C-c C-o") 'ink-follow-link-at-point)
-    (define-key map (kbd "C-c C-h") 'ink-display-manual)
-    ;; Visibility cycling
-    (define-key map (kbd "TAB") 'ink-cycle)
-    (define-key map (kbd "<S-iso-lefttab>") 'ink-shifttab)
-    (define-key map (kbd "<S-tab>")  'ink-shifttab)
-    (define-key map (kbd "<backtab>") 'ink-shifttab)
-    map)
-  "Keymap for ink major mode.")
-
-(defvar ink-mode-mouse-map
-  (let ((map (make-sparse-keymap)))
-    (define-key map [follow-link] 'mouse-face)
-    (define-key map [mouse-2] #'ink-follow-link-at-point)
-    map)
-  "Keymap for following links with mouse.")
+(defvar-keymap ink-mode-mouse-map
+  :doc "Mouse map for clickable ink links."
+  ;; This makes it look clickable
+  "<follow-link>" 'mouse-face
+  "<mouse-2>" #'ink-follow-link-at-point)
 
 (easy-menu-define ink-mode-menu ink-mode-map
   "Menu for `ink-mode'."
@@ -152,6 +143,32 @@ Group 2 matches a link text")
   "^\\s-*\\(TODO\\|//\\)"
   "Regexp identifying Ink comments.")
 
+
+;;; Compiler lookup
+
+(defcustom ink-inklecate-command "inklecate"
+  "Command used to invoke inklecate.
+Absolute path or a program used looked up in variable `exec-path'."
+  :group 'ink
+  :type '(choice (const :tag "inklecate (looked up in PATH)" "inklecate")
+                 (string :tag "Path or command name")))
+
+(defcustom ink-imenu-include-labels t
+  "If non-nil, include labels in the imenu index."
+  :group 'ink
+  :type 'boolean)
+
+(defun ink--inklecate-executable ()
+  "Return absolute path to inklecate or nil.
+Resolve the `ink-inklecate-command'.  Should be used whenever inklecate
+is invoked."
+  (when-let* ((cmd ink-inklecate-command))
+    (or
+     ;; Resolve absolute paths.
+     (and (file-name-absolute-p cmd) (file-executable-p cmd) cmd)
+     ;; Or check PATH.
+     (executable-find cmd))))
+
 
 ;;; Link following
 
@@ -293,7 +310,7 @@ to search."
                        (message "Jumping to %s" title)
                        (goto-char position)
                        (ignore-errors (outline-show-subtree)))
-          (user-error "Link `%s' not found. Is it in another file?" title))))))
+          (user-error "Link `%s' not found.  Is it in another file?" title))))))
 
 (defun ink-follow-file-link ()
   "Find file matching the link at point."
@@ -453,23 +470,23 @@ keyword."
   "If non-nil, force using spaces between choices and gathers.
 You'd get something like:
 
--   I looked at Monsieur Fogg
-    *   ... and I could contain myself no longer.
-        'What is the purpose of our journey, Monsieur?'
-        'A wager,' he replied.
-        * *   'A wager!'[] I returned.
-              He nodded.
-            * * *   'But surely that is foolishness!'
+  -   I looked at Monsieur Fogg
+      *   ... and I could contain myself no longer.
+          \\='What is the purpose of our journey, Monsieur?\\='
+          \\='A wager,\\=' he replied.
+          * *   \\='A wager!\\='[] I returned.
+                He nodded.
+              * * *   \\='But surely that is foolishness!\\='
 
 Otherwise, use the setting of `indent-tabs-mode', which may give:
 
--   I looked at Monsieur Fogg
-    *   ... and I could contain myself no longer.
-        'What is the purpose of our journey, Monsieur?'
-        'A wager,' he replied.
-        *   *   'A wager!'[] I returned.
-                He nodded.
-            *   *   *   'But surely that is foolishness!'"
+  -   I looked at Monsieur Fogg
+      *   ... and I could contain myself no longer.
+          \\='What is the purpose of our journey, Monsieur?\\='
+          \\='A wager,\\=' he replied.
+          *   *   \\='A wager!\\='[] I returned.
+                  He nodded.
+              *   *   *   \\='But surely that is foolishness!\\='"
   :group 'ink
   :type 'boolean)
 
@@ -490,7 +507,7 @@ Otherwise, use the setting of `indent-tabs-mode', which may give:
     (when follow-indentation-p (back-to-indentation))))
 
 (defun ink-count-choices ()
-  "Return number of choices or gathers in line."
+  "Return the number of choices or gather markers on the current line."
   (interactive)
   (let ((choices 0))
     (save-excursion
@@ -509,7 +526,7 @@ Otherwise, use the setting of `indent-tabs-mode', which may give:
     (make-string (max 0 (- tab-width 1)) ? )))
 
 (defun ink-indent-choices ()
-  "Indent choices and gathers: add indentations between symbols."
+  "Indent choice and gather markers by adding spacing between symbols."
   (interactive)
   (save-excursion
     (beginning-of-line)
@@ -639,8 +656,8 @@ Otherwise, use the setting of `indent-tabs-mode', which may give:
          ;; Conditions inside brackets - ...:
          ((and (looking-at "^\\s-*-.*:\\s-*")
                (or
-                (seq-contains indentation-list 'bracket)
-                (seq-contains indentation-list 'bracket-cond)))
+                (memq 'bracket indentation-list)
+                (memq 'bracket-cond indentation-list)))
           ;; Pop until previous bracket
           (while (not (or (eq 'bracket (nth 0 indentation-list))
                           (eq 'bracket-cond (nth 0 indentation-list))))
@@ -743,10 +760,10 @@ Otherwise, use the setting of `indent-tabs-mode', which may give:
     indentation))
 
 (defun ink-calculate-choice-indentation (element indentation-list indentation)
-  "Get the number of columns to indent choices and gathers.
-This depends on previous indentation, and settings. ELEMENT is
+  "Return the number of columns to indent choice and gather markers.
+This depends on previous indentation, and settings.  ELEMENT is
 the current element in the INDENTATION-LIST for the lign to
-indent. INDENTATION is the current sum."
+indent.  INDENTATION is the current sum."
   (let (value)
     (if ink-indent-choices-with-spaces
         (if (eq element (nth 0 indentation-list))
@@ -769,28 +786,18 @@ indent. INDENTATION is the current sum."
 
 ;;; Ink-play
 
-(defvar ink-play-mode-hook nil)
-
-(defvar ink-play-mode-map
-  (let ((map (make-sparse-keymap)))
-    (define-key map (kbd "C-c C-h") 'ink-display-manual)
-    (define-key map (kbd "C-c C-c") 'ink-play)
-    (define-key map (kbd "C-c C-p") 'ink-play-knot)
-    map)
-  "Keymap for ink-play mode.
-
+(defvar-keymap ink-play-mode-map
+  :doc "Keymap for `ink-play' mode.
 `ink-play' and `ink-play-knot' in this mode will re-run the last
-ink-play and ink-play-knot commands in the same buffer.")
+`ink-play' and `ink-play-knot' commands in the same buffer."
+  "C-c C-h" #'ink-display-manual
+  "C-c C-c" #'ink-play
+  "C-c C-p" #'ink-play-knot)
 
 (define-derived-mode ink-play-mode comint-mode "Ink-Play"
   "Major mode for `ink-play'.
-
-Derives from comint-mode, adds a few ink bindings.")
-
-(defcustom ink-inklecate-path (executable-find "inklecate")
-  "The path to the Inklecate executable."
-  :group 'ink
-  :type '(file))
+Derives from `comint-mode', adds a few ink bindings."
+  (add-hook 'comint-preoutput-filter-functions #'ink-comint-filter-output nil t))
 
 (defvar-local ink-comint-do-filter nil)
 
@@ -799,14 +806,15 @@ Derives from comint-mode, adds a few ink bindings.")
   (interactive)
   (ink-play t))
 
-;; last file-name and knot to support replaying
-(setq ink-last-played-file-name nil)
-(setq ink-last-played-knot nil)
+(defvar ink-last-played-file-name nil
+  "Last known file name that was run.")
+(defvar ink-last-played-knot nil
+  "Last known knot name that was run .")
 
 (defun ink-play (&optional go-to-knot)
   "Play the current ink buffer.
 If the GO-TO-KNOT optional argument is non-nil, start at the knot
-or stitch at point. In that case we issue \"-> knot.stitch\" to
+or stitch at point.  In that case we issue \"-> knot.stitch\" to
 the process, and suppress the beginning output using the comint
 output filter."
   (interactive "P")
@@ -836,13 +844,13 @@ output filter."
                ;; has finished, so kick it off again
                in-ink-play-buffer)
               (progn
-                (comint-exec "*Ink*" "Ink" ink-inklecate-path
+                (comint-exec "*Ink*" "Ink" (ink--inklecate-executable)
                              nil `("-p" ,file-name))
                 "*Ink*")
             ;; no *Ink* buffer yet, need to create it
             (progn
               (let ((new-buffer
-                     (make-comint "Ink" ink-inklecate-path nil
+                     (make-comint "Ink" (ink--inklecate-executable) nil
                                   "-p" (buffer-file-name))))
                 (set-buffer new-buffer)
                 (ink-play-mode)
@@ -873,12 +881,12 @@ output filter."
 
 (defun ink-filter-output-line (line)
   "Filter single line of text from Inklecate's output.
-The filter is active only on starting play. It outputs all
+The filter is active only on starting play.  It outputs all
 errors, warnings and infos appearing in LINE, and discards the
 rest."
   (let ((result ""))
     (if ink-comint-do-filter
-        (cond ((string-match-p "^\\(ERROR:\\|WARNING:\\|TODO\\)" line)
+        (cond ((string-match-p "^\\(ERROR:\\|WARNING:\\|TODO:\\)" line)
                (setq result (concat line "\n")))
               ((string-match-p "\\?>" line)
                (setq result (concat (substring line 3) "\n"))
@@ -897,113 +905,133 @@ directly at a knot... OUTPUT is the output to be filtered."
       (setq output (mapconcat #'ink-filter-output-line (split-string output "\n") "")))
   output)
 
-(add-hook 'comint-preoutput-filter-functions #'ink-comint-filter-output)
-
 
-;;; Error checking with flymake
+;;; Error checking with Flymake
+
+(defun ink--flymake-type-from-tag (tag)
+  "Map compiler error TAG to Flymake warning types."
+  (pcase (upcase tag)
+    ("ERROR" :error)
+    ("WARNING" :warning)
+    ("TODO" :note)
+    ;; NOTE: shouldn't happen as we use a regexp to catch all options
+    (_ (error "Uknown error tag"))))
+
+(defun ink--flymake-parse-line (line)
+  "Parse a single inklecate output LINE.
+Return a plist (:file :line :type :msg) or nil."
+  (when (string-match
+         ;; Error tag here should match the ones in
+         ;; `ink--flymake-type-from-tag'.
+         (rx bol
+             ;; Tag
+             (group (or "ERROR" "WARNING" "TODO")) ":" (* space)
+             ;; File
+             (? "'") (group (+ (not (any "'")))) (? "'")
+             ;; Line
+             (* space) "line" (+ space) (group (+ digit)) ":" (* space)
+             ;; End of line
+             (group (* any)) eol)
+         line)
+    (list :file (match-string 2 line)
+          :line (string-to-number (match-string 3 line))
+          :type (ink--flymake-type-from-tag (match-string 1 line))
+          :msg  (string-trim (match-string 4 line)))))
+
+(defun ink--flymake-parse-output (output)
+  "Parse inklecate OUTPUT into a list of plists."
+  (delq nil (mapcar #'ink--flymake-parse-line (split-string output "\n" t))))
+
+(defun ink--flymake-same-file-p (a b default-dir)
+  "Non-nil if files A and B refer to the same file (best-effort).
+A or B may be relative; DEFAULT-DIR is used to resolve."
+  (let ((aa (expand-file-name a default-dir))
+        (bb (expand-file-name b default-dir)))
+    (or (file-equal-p aa bb)
+        ;; Some inklecate outputs only basenames; keep this as a fallback.
+        (string-equal (file-name-nondirectory aa)
+                      (file-name-nondirectory bb)))))
+
+(defun ink--flymake-items-to-diagnostics (items source-buf temp-file default-dir)
+  "Convert parsed ITEMS into Flymake diagnostics for SOURCE-BUF.
+TEMP-FILE - temporary file used to run the checks.  DEFAULT-DIR -
+directory used to run the check."
+  (let (diags)
+    (dolist (it items)
+      (let* ((f (plist-get it :file))
+             (line (or (plist-get it :line) 1))
+             (type (or (plist-get it :type) :warning))
+             (msg (or (plist-get it :msg) "")))
+        (when (ink--flymake-same-file-p f temp-file default-dir)
+          (pcase-let ((`(,beg . ,end) (flymake-diag-region source-buf line)))
+            (push (flymake-make-diagnostic source-buf beg end type msg) diags)))))
+    diags))
 
 (defvar-local ink--flymake-proc nil)
 
-(defun ink-flymake (report-fn &rest _args)
-  "Ink backend for Flymake.
-Creates temporary files and passes their names as arguments to
-`ink-inklecate-path' (which see). The output of this command is
-analyzed for error and warning messages."
-  (unless (executable-find ink-inklecate-path)
-    (error "Cannot find a suitable checker"))
-  ;; If a live process launched in an earlier check was found, that
-  ;; process is killed.  When that process's sentinel eventually runs,
-  ;; it will notice its obsoletion, since it have since reset
-  ;; `ink-flymake-proc' to a different value
-  ;;
+;;; TODO: Also report errors in INCLUDE'd files.
+(defun ink-flymake-inklecate (report-fn &rest _args)
+  "Flymake backend for Ink using inklecate.
+REPORT-FN - Flymake diagnostics reporting function."
+  ;; Check if the compiler is available, disable otherwise.
+  (unless (ink--inklecate-executable)
+    (error "Unable to find inklecape"))
+
+  ;; Process stuck?
   (when (process-live-p ink--flymake-proc)
     (kill-process ink--flymake-proc))
-  ;; From ‘flymake-proc-init-create-temp-buffer-copy’.
-  ;;
+
   (let* ((source (current-buffer))
-         (temp-file   (flymake-proc-init-create-temp-buffer-copy
-                       'flymake-proc-create-temp-inplace))
-         (local-file  (file-relative-name
-                       temp-file
-                       (file-name-directory buffer-file-name)))
-         (json-file   (concat local-file ".json")))
-    ;; Save the current buffer, the narrowing restriction, remove any
-    ;; narrowing restriction.
-    ;;
+         (source-file (buffer-file-name source))
+         (default-dir (file-name-directory source-file))
+
+         ;; Tmp source file in same dir so INCLUDE relative paths
+         ;; work.
+         (temp-ink (make-temp-file
+                    (expand-file-name
+                     (concat (file-name-base source-file) "-flymake-")
+                     default-dir)
+                    nil ".ink"))
+
+         ;; Dump stdout somewhere.
+         (cmd (append (list (ink--inklecate-executable)
+                            "-o" (or null-device (make-temp-name "NUL")))
+                      (list temp-ink))))
+
+    ;; Write to a temporary file.
     (save-restriction
       (widen)
-      ;; Reset the `ink--flymake-proc' process to a new process
-      ;; calling the ink tool.
-      ;;
-      (setq ink--flymake-proc
-            (make-process
-             :name "ink-flymake" :noquery t :connection-type 'pipe
-             ;; Make output go to a temporary buffer.
-             ;;
-             :buffer (generate-new-buffer " *ink-flymake*")
-             :command (list ink-inklecate-path "-o" json-file local-file)
-             :sentinel
-             (lambda (proc _event)
-               ;; Check that the process has indeed exited, as it might
-               ;; be simply suspended.
-               ;;
-               (when (eq 'exit (process-status proc))
-                 (unwind-protect
-                     ;; Only proceed if `proc' is the same as
-                     ;; `ink--flymake-proc', which indicates that
-                     ;; `proc' is not an obsolete process.
-                     ;;
-                     (if (with-current-buffer source (eq proc ink--flymake-proc))
-                         (with-current-buffer (process-buffer proc)
-                           (goto-char (point-min))
-                           ;; Parse the output buffer for diagnostic's
-                           ;; messages and locations, collect them in a list
-                           ;; of objects, and call `report-fn'.
-                           ;;
-                           (cl-loop
-                            while (search-forward-regexp
-                                   "^\\(.*?\\): '.*?' line \\([0-9]+\\): \\(.*\\)$"
-                                   nil t)
-                            for msg = (match-string 3)
-                            for (beg . end) = (flymake-diag-region
-                                               source
-                                               (string-to-number (match-string 2)))
-                            for type = (cond
-                                        ((string-match "ERROR" (match-string 1)) :error)
-                                        ((string-match "WARNING" (match-string 1)) :warning)
-                                        ((string-match "TODO" (match-string 1)) :note)
-                                        (t :warning))
-                            collect (flymake-make-diagnostic source
-                                                             beg
-                                                             end
-                                                             type
-                                                             msg)
-                            into diags
-                            finally (funcall report-fn diags)))
-                       ;; Cleanup the temporary buffer used to hold the
-                       ;; check's output.
-                       ;;
-                       (kill-buffer (process-buffer proc)))
-                   ;; Delete temporary files
-                   (flymake-proc--safe-delete-file local-file)
-                   (flymake-proc--safe-delete-file json-file))))))
-      ;; Send the buffer contents to the process's stdin, followed by
-      ;; an EOF.
-      ;;
-      (process-send-region ink--flymake-proc (point-min) (point-max))
-      (process-send-eof ink--flymake-proc))))
+      (write-region (point-min) (point-max) temp-ink nil 'silent))
+
+    ;; Launch the process.
+    (setq ink--flymake-proc
+          (make-process
+           :name "ink-flymake"
+           :noquery t
+           :buffer (generate-new-buffer " *ink-flymake*")
+           :command cmd
+           :sentinel
+           (lambda (proc _event)
+             (when (memq (process-status proc) '(exit signal))
+               (unwind-protect
+                   (when (and (buffer-live-p source)
+                              (eq proc ink--flymake-proc))
+                     ;; buffer string -> parse -> diags -> report.
+                     (funcall
+                      report-fn
+                      (thread-first
+                        (with-current-buffer (process-buffer proc)
+                          (buffer-string))
+                        (ink--flymake-parse-output)
+                        (ink--flymake-items-to-diagnostics
+                         source temp-ink default-dir))))
+                 ;; Cleanup on exit/failure.
+                 (ignore-errors (delete-file temp-ink))
+                 (kill-buffer (process-buffer proc)))))))))
 
 
 ;;; Outline
 
-;; Outline functions were derived from markdown-mode.el, in turn
-;; originally derived from from org.el.
-
-(defvar-local ink-cycle-global-status 1)
-(defvar-local ink-cycle-subtree-status nil)
-
-(defalias 'ink-end-of-heading 'outline-end-of-heading)
-
 (defun ink-end-of-subtree (&optional invisible-OK)
   "Move to the end of the current subtree.
 Only visible heading lines are considered, unless INVISIBLE-OK is
@@ -1058,109 +1086,83 @@ stitch."
           (setq title-list (append title-list (reverse (split-string knot-name "\\.")))))
       title-list)))
 
-(defun ink-shifttab ()
-  "S-TAB keybinding: cycle global heading visibility by calling `ink-cycle' with argument t."
-  (interactive)
-  (ink-cycle t))
-
-(defun ink-cycle (&optional arg)
-  "Visibility cycling for Ink mode.
-If ARG is t, perform global visibility cycling. If the point is
-at a header, cycle visibility of the corresponding subtree.
-Otherwise, indent the current line or insert a tab, as
-appropriate, by calling `indent-for-tab-command'."
-  (interactive "P")
-  (cond
-   ;; Global cycling
-   ((eq arg t)
-    (cond
-     ;; Move from overview to contents
-     ((and (eq last-command this-command)
-           (eq ink-cycle-global-status 2))
-      (outline-hide-sublevels 1)
-      (message "CONTENTS")
-      (setq ink-cycle-global-status 3))
-     ;; Move from contents to all
-     ((and (eq last-command this-command)
-           (eq ink-cycle-global-status 3))
-      (outline-show-all)
-      (message "SHOW ALL")
-      (setq ink-cycle-global-status 1))
-     ;; Defaults to overview
-     (t
-      (outline-hide-body)
-      (message "OVERVIEW")
-      (setq ink-cycle-global-status 2))))
-
-   ;; At a heading: rotate between three different views
-   ((thing-at-point-looking-at ink-regex-header)
-    (outline-back-to-heading)
-    (let (eoh eol eos)
-      ;; Determine boundaries
-      (save-excursion
-        (outline-back-to-heading)
-        (save-excursion
-          (beginning-of-line 2)
-          (while (and (not (eobp)) ;; this is like `next-line'
-                      (get-char-property (1- (point)) 'invisible))
-            (beginning-of-line 2)) (setq eol (point)))
-        (ink-end-of-heading)   (setq eoh (point))
-        (ink-end-of-subtree t)
-        (skip-chars-forward " \t\n")
-        (beginning-of-line 1) ; in case this is an item
-        (setq eos (1- (point))))
-      ;; Find out what to do next and set `this-command'
-      (cond
-       ;; Nothing is hidden behind this heading
-       ((= eos eoh)
-        (message "EMPTY ENTRY")
-        (setq ink-cycle-subtree-status nil))
-       ;; Entire subtree is hidden in one line: open it
-       ((>= eol eos)
-        ;; (ink-show-entry)
-        (outline-show-entry)
-        (outline-show-children)
-        (message "CHILDREN")
-        (setq ink-cycle-subtree-status 'children))
-       ;; We just showed the children, now show everything.
-       ((and (eq last-command this-command)
-             (eq ink-cycle-subtree-status 'children))
-        (outline-show-subtree)
-        (message "SUBTREE")
-        (setq ink-cycle-subtree-status 'subtree))
-       ;; Default action: hide the subtree.
-       (t
-        (outline-hide-subtree)
-        (message "FOLDED")
-        (setq ink-cycle-subtree-status 'folded)))))
-
-   ;; Otherwise, indent as appropriate
-   (t
-    (indent-for-tab-command))))
-
 (defun ink-outline-level ()
   "Return the depth to which a statement is nested in the outline."
-  (if (> (length (match-string-no-properties 1)) 1)
-      1
-    2))
+  (if (> (length (match-string-no-properties 1)) 1) 1 2))
+
+
+;;; Symbols
+
+(defun ink--collect-symbols ()
+  "Return parsed symbols from the current buffer.
+Each result entry is a plist with `:kind', `:name', `:position'
+and `:completion-targets' keys."
+  (let (symbols)
+    (save-excursion
+      (goto-char (point-min))
+      (while (not (eobp))
+        (beginning-of-line)
+        (cond
+         ((looking-at ink-regex-header)
+          (when-let* ((kind (if (= (ink-outline-level) 1) 'knot 'stitch))
+                      (short-name (match-string-no-properties 3))
+                      (full-name (ink-get-knot-name)))
+            (push (list :kind kind
+                        :name full-name
+                        :position (line-beginning-position)
+                        :completion-targets (delete-dups (list short-name full-name)))
+                  symbols)))
+         ((looking-at ink-regex-label)
+          (when-let* ((label-parts (ink-get-label-name))
+                      (short-name (car label-parts))
+                      (full-name (mapconcat #'identity (reverse label-parts) ".")))
+            (push (list :kind 'label
+                        :name full-name
+                        :position (line-beginning-position)
+                        :completion-targets (delete-dups (list short-name full-name)))
+                  symbols))))
+        (forward-line 1)))
+    (nreverse symbols)))
+
+(defun ink--imenu-create-index ()
+  "Return imenu index for the current Ink buffer."
+  (let (knots stitches labels)
+    (dolist (symbol (ink--collect-symbols))
+      (let ((entry (cons (plist-get symbol :name)
+                         (plist-get symbol :position))))
+        (pcase (plist-get symbol :kind)
+          ('knot (push entry knots))
+          ('stitch (push entry stitches))
+          ('label (push entry labels)))))
+    (setq knots (nreverse knots))
+    (setq stitches (nreverse stitches))
+    (setq labels (nreverse labels))
+    (delq nil (list (and knots (cons "Knots" knots))
+                    (and stitches (cons "Stitches" stitches))
+                    (and ink-imenu-include-labels labels (cons "Labels" labels))))))
+
+(defun ink--current-symbol-name ()
+  "Return current Ink symbol name at point, or nil if unknown.
+When `ink-imenu-include-labels' is non-nil, labels can be returned."
+  (let ((pos (point))
+        name)
+    (dolist (symbol (ink--collect-symbols))
+      (let ((kind (plist-get symbol :kind)))
+        (when (and (<= (plist-get symbol :position) pos)
+                   (or (not (eq kind 'label))
+                       ink-imenu-include-labels))
+          (setq name (plist-get symbol :name)))))
+    name))
 
 
 ;;; Autocomplete
 
 (defun ink-get-headers-and-labels ()
-  "Return a list of all header and label hierarchies."
-  (let ((headers-labels (list "END" "DONE")) match)
-    (save-excursion
-      (goto-char (point-min))
-      (while (re-search-forward ink-regex-header nil t)
-        (when (setq match (match-string-no-properties 3))
-          (push match headers-labels)
-          (push (ink-get-knot-name) headers-labels)))
-      (goto-char (point-min))
-      (while (re-search-forward ink-regex-label nil t)
-        (when (setq match (match-string-no-properties 2))
-          (push match headers-labels)
-          (push (concat (ink-get-knot-name) "." match) headers-labels))))
+  "Return divert completion targets for headers and labels."
+  (let ((headers-labels (list "END" "DONE")))
+    (dolist (symbol (ink--collect-symbols))
+      (dolist (target (plist-get symbol :completion-targets))
+        (push target headers-labels)))
     (sort (delete-dups headers-labels) #'string<)))
 
 (defun ink-completion-at-point ()
@@ -1188,19 +1190,24 @@ Completion is only provided for diverts."
 
 ;;;###autoload
 (defun ink-load-snippets()
-  "Load snippets if yasnippet installed and `ink-snippet-dir' is set."
+  "Manually load snippets if yasnippet installed and `ink-snippet-dir' is set.
+This command is explicit and can be used to reload bundled snippets."
   (interactive)
   (when ink-snippet-dir
     (cond
      ((fboundp 'yas-load-directory)
       (yas-load-directory ink-snippet-dir))
      ((fboundp 'yas/load-directory)
-      (yas/load-directory ink-snippet-dir)))))
+      (yas/load-directory ink-snippet-dir))
+     (t
+      (user-error
+       "YASnippet is not loaded yet; Ink snippets were not loaded")))))
 
 
 ;;; Help
 
 (defun ink-display-manual ()
+  "Display the Writing With Ink manual."
   (interactive)
   (info-display-manual "WritingWithInk"))
 
@@ -1209,9 +1216,7 @@ Completion is only provided for diverts."
 
 ;;;###autoload
 (define-derived-mode ink-mode prog-mode "Ink"
-  "Major mode for editing interactive fiction using the Ink
-  scripting language."
-  :syntax-table ink-mode-syntax-table
+  "Major mode for the Ink interactive fiction scripting language."
 
   ;; Syntax
   (setq-local comment-start "// ")
@@ -1219,7 +1224,7 @@ Completion is only provided for diverts."
   (setq-local comment-use-syntax t)
   (setq-local comment-end "")
   (setq-local comment-auto-fill-only-comments t)
-  (setq font-lock-defaults '(ink-font-lock-keywords))
+  (setq-local font-lock-defaults '(ink-font-lock-keywords))
 
   ;; Indent
   (setq-local paragraph-ignore-fill-prefix t)
@@ -1233,14 +1238,14 @@ Completion is only provided for diverts."
   ;; Outline
   (setq-local outline-regexp ink-regex-header)
   (setq-local outline-level #'ink-outline-level)
-  ;; Cause use of ellipses for invisible text.
-  (add-to-invisibility-spec '(outline . t))
+
+  ;; Imenu
+  (setq-local imenu-create-index-function #'ink--imenu-create-index)
+  (add-hook 'which-func-functions #'ink--current-symbol-name nil 'local)
+  (setq-local add-log-current-defun-function #'ink--current-symbol-name)
 
   ;; Flymake
-  (add-hook 'flymake-diagnostic-functions 'ink-flymake nil t)
-
-  ;; Snippets
-  (ink-load-snippets))
+  (add-hook 'flymake-diagnostic-functions 'ink-flymake-inklecate nil t))
 
 ;;;###autoload
 (add-to-list 'auto-mode-alist '("\\.ink\\'" . ink-mode))
